######################################################################
#
# File: test/unit/console_tool/test_notification_rules.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import json

import pytest


@pytest.fixture()
def bucket_notification_rule(b2_cli, bucket):
    rule = {
        "eventTypes": ["b2:ObjectCreated:*"],
        "isEnabled": True,
        "isSuspended": False,
        "name": "test-rule",
        "objectNamePrefix": "",
        "suspensionReason": "",
        "targetConfiguration": {
            "targetType": "webhook",
            "url": "https://example.com/webhook",
        },
    }
    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "create",
            "--json",
            f"b2://{bucket}",
            "test-rule",
            "--webhook-url",
            "https://example.com/webhook",
            "--event-type",
            "b2:ObjectCreated:*",
        ],
    )
    actual_rule = json.loads(stdout)
    assert actual_rule == rule
    return actual_rule


def test_notification_rules__list_all(b2_cli, bucket, bucket_notification_rule):
    _, stdout, _ = b2_cli.run([
        "notification-rules",
        "list",
        f"b2://{bucket}",
    ])
    assert (
        stdout == f"""\
Notification rules for b2://{bucket}/ :
- name: test-rule
  eventTypes:
    - b2:ObjectCreated:*
  isEnabled: true
  isSuspended: false
  objectNamePrefix: ''
  suspensionReason: ''
  targetConfiguration:
    targetType: webhook
    url: https://example.com/webhook
"""
    )


def test_notification_rules__list_all_json(b2_cli, bucket, bucket_notification_rule):
    _, stdout, _ = b2_cli.run([
        "notification-rules",
        "list",
        "--json",
        f"b2://{bucket}",
    ])
    assert json.loads(stdout) == [bucket_notification_rule]


def test_notification_rules__update(b2_cli, bucket, bucket_notification_rule):
    bucket_notification_rule["isEnabled"] = False
    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "update",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
            "--disable",
            "--custom-header",
            "X-Custom-Header=value=1",
        ],
    )
    bucket_notification_rule["targetConfiguration"]["customHeaders"] = {
        "X-Custom-Header": "value=1"
    }
    assert json.loads(stdout) == bucket_notification_rule


def test_notification_rules__update__no_such_rule(b2_cli, bucket, bucket_notification_rule):
    b2_cli.run(
        [
            "notification-rules",
            "update",
            f"b2://{bucket}",
            f'{bucket_notification_rule["name"]}-unexisting',
            "--disable",
        ],
        expected_stderr=(
            "ERROR: rule with name 'test-rule-unexisting' does not exist on bucket "
            "'my-bucket', available rules: ['test-rule']\n"
        ),
        expected_status=1,
    )


def test_notification_rules__update__custom_header_malformed(
    b2_cli, bucket, bucket_notification_rule
):
    bucket_notification_rule["isEnabled"] = False
    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "update",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
            "--disable",
            "--custom-header",
            "X-Custom-Header: value",
        ],
    )
    bucket_notification_rule["targetConfiguration"]["customHeaders"] = {
        "X-Custom-Header: value": ""
    }
    assert json.loads(stdout) == bucket_notification_rule


def test_notification_rules__delete(b2_cli, bucket, bucket_notification_rule):
    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "delete",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
        ],
    )
    assert stdout == "Rule 'test-rule' has been deleted from b2://my-bucket/\n"


def test_notification_rules__delete_no_such_rule(b2_cli, bucket, bucket_notification_rule):
    b2_cli.run(
        [
            "notification-rules",
            "delete",
            f"b2://{bucket}",
            f'{bucket_notification_rule["name"]}-unexisting',
        ],
        expected_stderr=(
            "ERROR: no such rule to delete: 'test-rule-unexisting', available rules: ['test-rule'];"
            " No rules have been deleted.\n"
        ),
        expected_status=1,
    )


@pytest.mark.parametrize(
    "args,expected_stdout",
    [
        (["-q"], ""),
        ([], "No notification rules for b2://my-bucket/\n"),
        (["--json"], "[]\n"),
    ],
)
def test_notification_rules__no_rules(b2_cli, bucket, args, expected_stdout):
    b2_cli.run(
        ["notification-rules", "list", f"b2://{bucket}", *args],
        expected_stdout=expected_stdout,
    )


def test_notification_rules__disable_enable(b2_cli, bucket, bucket_notification_rule):
    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "disable",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
        ],
    )
    assert json.loads(stdout) == {**bucket_notification_rule, "isEnabled": False}

    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "enable",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
        ],
    )
    assert json.loads(stdout) == {**bucket_notification_rule, "isEnabled": True}


@pytest.mark.parametrize(
    "command",
    ["disable", "enable"],
)
def test_notification_rules__disable_enable__no_such_rule(
    b2_cli, bucket, bucket_notification_rule, command
):
    b2_cli.run(
        [
            "notification-rules",
            command,
            f"b2://{bucket}",
            f'{bucket_notification_rule["name"]}-unexisting',
        ],
        expected_stderr=(
            "ERROR: rule with name 'test-rule-unexisting' does not exist on bucket "
            "'my-bucket', available rules: ['test-rule']\n"
        ),
        expected_status=1,
    )


def test_notification_rules__sign_secret(b2_cli, bucket, bucket_notification_rule):
    b2_cli.run(
        [
            "notification-rules",
            "update",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
            "--sign-secret",
            "new-secret",
        ],
        expected_status=2,
    )

    _, stdout, _ = b2_cli.run(
        [
            "notification-rules",
            "update",
            "--json",
            f"b2://{bucket}",
            bucket_notification_rule["name"],
            "--sign-secret",
            "7" * 32,
        ],
    )
    bucket_notification_rule["targetConfiguration"]["hmacSha256SigningSecret"] = "7" * 32
    assert json.loads(stdout) == bucket_notification_rule

    assert json.loads(b2_cli.run(["notification-rules", "list", "--json", f"b2://{bucket}"],)[1]
                     ) == [bucket_notification_rule]



######################################################################
#
# File: test/static/test_licenses.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from datetime import datetime
from glob import glob
from itertools import islice

import pytest

FIXER_CMD = "python test/static/test_licenses.py"
LICENSE_HEADER_TMPL = """\
######################################################################
#
# File: {path}
#
# Copyright {year} Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
"""


def get_file_header_errors(file_path_glob: str) -> dict[str, str]:
    failed_files = {}
    for file in glob(file_path_glob, recursive=True):
        if file.startswith('build/'):
            # built files naturally have a different file path than source files
            continue
        with open(file) as fd:
            file = file.replace(
                '\\', '/'
            )  # glob('**/*.py') on Windows returns "b2\console_tool.py" (wrong slash)
            head = ''.join(islice(fd, 9))
            if 'All Rights Reserved' not in head:
                failed_files[file] = 'Missing "All Rights Reserved" in the header'
            elif file not in head:
                failed_files[file] = 'Wrong file name in the header'
    return failed_files


def test_files_headers():
    failed_files = get_file_header_errors('**/*.py')
    if failed_files:
        error_msg = '; '.join(f'{path}:{error}' for path, error in failed_files.items())
        pytest.fail(f'Bad file headers in files (you may want to run {FIXER_CMD!r}): {error_msg}')


def insert_header(file_path: str):
    with open(file_path, 'r+') as fd:
        content = fd.read()
        fd.seek(0)
        fd.write(LICENSE_HEADER_TMPL.format(
            path=file_path,
            year=datetime.now().year,
        ))
        fd.write(content)


def _main():
    failed_files = get_file_header_errors('**/*.py')
    for filepath in failed_files:
        insert_header(filepath)


if __name__ == '__main__':
    _main()



######################################################################
#
# File: test/unit/test_apiver.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import unittest

import pytest


@pytest.fixture
def inject_apiver_int(request, apiver_int):
    request.cls.apiver_int = apiver_int


@pytest.mark.usefixtures('inject_apiver_int')
class UnitTestClass(unittest.TestCase):
    apiver_int: int

    @pytest.mark.apiver(to_ver=3)
    def test_passes_below_and_on_v3(self):
        assert self.apiver_int <= 3

    @pytest.mark.apiver(from_ver=4)
    def test_passes_above_and_on_v4(self):
        assert self.apiver_int >= 4

    @pytest.mark.apiver(3)
    def test_passes_only_on_v3(self):
        assert self.apiver_int == 3

    @pytest.mark.apiver(4)
    def test_passes_only_on_v4(self):
        assert self.apiver_int == 4

    @pytest.mark.apiver(3, 4)
    def test_passes_on_both_v3_and_v4(self):
        assert self.apiver_int in {3, 4}


@pytest.mark.apiver(to_ver=3)
def test_passes_below_and_on_v3(apiver_int):
    assert apiver_int <= 3


@pytest.mark.apiver(from_ver=4)
def test_passes_above_and_on_v4(apiver_int):
    assert apiver_int >= 4


@pytest.mark.apiver(3)
def test_passes_only_on_v3(apiver_int):
    assert apiver_int == 3


@pytest.mark.apiver(4)
def test_passes_only_on_v4(apiver_int):
    assert apiver_int == 4


@pytest.mark.apiver(3, 4)
def test_passes_on_both_v3_and_v4(apiver_int):
    assert apiver_int in {3, 4}



######################################################################
#
# File: test/integration/conftest.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging
import os
import pathlib
import re
import subprocess
import sys
import tempfile
from os import environ, path
from tempfile import TemporaryDirectory

import pytest
from b2sdk.v2 import B2_ACCOUNT_INFO_ENV_VAR, XDG_CONFIG_HOME_ENV_VAR, Bucket

from b2._internal.version_listing import (
    CLI_VERSIONS,
    LATEST_STABLE_VERSION,
    UNSTABLE_CLI_VERSION,
    get_int_version,
)

from ..helpers import b2_uri_args_v3, b2_uri_args_v4
from .helpers import NODE_DESCRIPTION, RNG_SEED, Api, CommandLine, bucket_name_part, random_token

logger = logging.getLogger(__name__)

GENERAL_BUCKET_NAME_PREFIX = 'clitst'
TEMPDIR = tempfile.gettempdir()
ROOT_PATH = pathlib.Path(__file__).parent.parent.parent


@pytest.fixture(scope='session', autouse=True)
def summary_notes(request, worker_id):
    capmanager = request.config.pluginmanager.getplugin("capturemanager")
    with capmanager.global_and_fixture_disabled():
        log_handler = logging.StreamHandler(sys.stderr)
    log_fmt = logging.Formatter(f'{worker_id} %(asctime)s %(levelname).1s %(message)s')
    log_handler.setFormatter(log_fmt)
    logger.addHandler(log_handler)

    class Notes:
        def append(self, note):
            logger.info(note)

    return Notes()


@pytest.fixture(scope='session', autouse=True)
def node_stats(summary_notes):
    summary_notes.append(f"NODE={NODE_DESCRIPTION} seed={RNG_SEED}")


@pytest.hookimpl
def pytest_addoption(parser):
    parser.addoption(
        '--sut',
        default=f'{sys.executable} -m b2._internal.{UNSTABLE_CLI_VERSION}',
        help='Path to the System Under Test',
    )
    parser.addoption(
        '--env-file-cmd-placeholder',
        default=None,
        help=(
            'If specified, all occurrences of this string in `--sut` will be substituted with a '
            'path to a tmp file containing env vars to be used when running commands in tests. Useful '
            'for docker.'
        )
    )
    parser.addoption(
        '--as_version',
        default=None,
        help='Force running tests as a particular version of the CLI, '
        'useful if version cannot be determined easily from the executable',
    )
    parser.addoption('--cleanup', action='store_true', help='Perform full cleanup at exit')


def get_raw_cli_int_version(config) -> int | None:
    forced_version = config.getoption('--as_version')
    if forced_version:
        return int(forced_version)

    executable = config.getoption('--sut')
    # If the executable contains anything that looks like a proper version, we can try to pick it up.
    versions_list = '|'.join(CLI_VERSIONS)
    versions_match = re.search(rf'({versions_list})', executable)
    if versions_match:
        return get_int_version(versions_match.group(1))

    return None


def get_cli_int_version(config) -> int:
    return get_raw_cli_int_version(config) or get_int_version(LATEST_STABLE_VERSION)


@pytest.fixture(scope='session')
def apiver_int(request):
    return get_cli_int_version(request.config)


@pytest.fixture(scope='session')
def apiver(apiver_int):
    return f"v{apiver_int}"


@pytest.hookimpl
def pytest_report_header(config):
    cli_version = get_cli_int_version(config)
    return f'b2 apiver: {cli_version}'


@pytest.fixture(scope='session')
def cli_version(request) -> str:
    """
    Get CLI version name, i.e. b2v3, _b2v4, etc.
    """
    # The default stable version could be provided directly as e.g.: b2v3, but also indirectly as b2.
    # In case there is no direct version, we return the default binary name instead.
    raw_cli_version = get_raw_cli_int_version(request.config)
    if raw_cli_version is None:
        return 'b2'

    for version in CLI_VERSIONS:
        if get_int_version(version) == raw_cli_version:
            return version
    raise pytest.UsageError(f'Unknown CLI version: {raw_cli_version}')


@pytest.fixture(scope='session')
def application_key() -> str:
    key = environ.get('B2_TEST_APPLICATION_KEY')
    assert application_key, 'B2_TEST_APPLICATION_KEY is not set'
    yield key


@pytest.fixture(scope='session')
def application_key_id() -> str:
    key_id = environ.get('B2_TEST_APPLICATION_KEY_ID')
    assert key_id, 'B2_TEST_APPLICATION_KEY_ID is not set'
    yield key_id


@pytest.fixture(scope='session')
def realm() -> str:
    yield environ.get('B2_TEST_ENVIRONMENT', 'production')


@pytest.fixture
def bucket(bucket_factory) -> Bucket:
    return bucket_factory()


@pytest.fixture
def bucket_factory(b2_api, schedule_bucket_cleanup):
    def create_bucket(**kwargs):
        new_bucket = b2_api.create_bucket(**kwargs)
        schedule_bucket_cleanup(new_bucket.name, new_bucket.bucket_dict)
        return new_bucket

    yield create_bucket


@pytest.fixture(scope='function')
def bucket_name(bucket) -> str:
    yield bucket.name


@pytest.fixture(scope='function')
def file_name(bucket) -> str:
    file_ = bucket.upload_bytes(b'test_file', f'{random_token(8)}.txt')
    yield file_.file_name


@pytest.fixture(scope='function')  # , autouse=True)
def debug_print_buckets(b2_api):
    print('-' * 30)
    print('Buckets before test ' + environ['PYTEST_CURRENT_TEST'])
    num_buckets = b2_api.count_and_print_buckets()
    print('-' * 30)
    try:
        yield
    finally:
        print('-' * 30)
        print('Buckets after test ' + environ['PYTEST_CURRENT_TEST'])
        delta = b2_api.count_and_print_buckets() - num_buckets
        print(f'DELTA: {delta}')
        print('-' * 30)


@pytest.fixture(scope='session')
def this_run_bucket_name_prefix() -> str:
    yield GENERAL_BUCKET_NAME_PREFIX + bucket_name_part(8)


@pytest.fixture(scope='session')
def monkeysession():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope='session', autouse=True)
def auto_change_account_info_dir(monkeysession) -> str:
    """
    Automatically for the whole testing:
    1) temporary remove B2_APPLICATION_KEY and B2_APPLICATION_KEY_ID from environment
    2) create a temporary directory for storing account info database
    3) set B2_ACCOUNT_INFO_ENV_VAR to point to the temporary account info file
    """

    monkeysession.delenv('B2_APPLICATION_KEY_ID', raising=False)
    monkeysession.delenv('B2_APPLICATION_KEY', raising=False)

    # make b2sdk use temp dir for storing default & per-profile account information
    with TemporaryDirectory() as temp_dir:
        monkeysession.setenv(B2_ACCOUNT_INFO_ENV_VAR, path.join(temp_dir, '.b2_account_info'))
        monkeysession.setenv(XDG_CONFIG_HOME_ENV_VAR, temp_dir)
        yield temp_dir


@pytest.fixture(scope='session')
def b2_api(
    application_key_id,
    application_key,
    realm,
    this_run_bucket_name_prefix,
    auto_change_account_info_dir,
    summary_notes,
) -> Api:
    api = Api(
        application_key_id,
        application_key,
        realm,
        general_bucket_name_prefix=GENERAL_BUCKET_NAME_PREFIX,
        this_run_bucket_name_prefix=this_run_bucket_name_prefix,
    )
    yield api
    api.clean_buckets()
    # showing account_id in the logs is safe; so we explicitly prevent it from being redacted
    summary_notes.append(f"B2 Account ID: {api.account_id[:1]!r}{api.account_id[1:]!r}")
    summary_notes.append(f"Buckets names used during this tests: {api.bucket_name_log!r}")


@pytest.fixture(scope='module')
def global_b2_tool(
    request, application_key_id, application_key, realm, this_run_bucket_name_prefix, b2_api,
    auto_change_account_info_dir, b2_uri_args
) -> CommandLine:
    tool = CommandLine(
        request.config.getoption('--sut'),
        application_key_id,
        application_key,
        realm,
        this_run_bucket_name_prefix,
        request.config.getoption('--env-file-cmd-placeholder'),
        api_wrapper=b2_api,
        b2_uri_args=b2_uri_args,
    )
    tool.reauthorize(check_key_capabilities=True)  # reauthorize for the first time (with check)
    yield tool


@pytest.fixture(scope='function')
def b2_tool(global_b2_tool):
    """Automatically reauthorized b2_tool for each test (without check)"""
    global_b2_tool.reauthorize(check_key_capabilities=False)
    return global_b2_tool


@pytest.fixture
def account_info_file() -> pathlib.Path:
    return pathlib.Path(os.environ[B2_ACCOUNT_INFO_ENV_VAR]).expanduser()


@pytest.fixture
def schedule_bucket_cleanup(global_b2_tool):
    """
    Explicitly ask for buckets cleanup after the test

    This should be only used when testing `create-bucket` command; otherwise use `bucket_factory` fixture.
    """
    buckets_to_clean = {}

    def add_bucket_to_cleanup(bucket_name, bucket_dict: dict | None = None):
        buckets_to_clean[bucket_name] = bucket_dict

    yield add_bucket_to_cleanup
    if buckets_to_clean:
        global_b2_tool.reauthorize(
            check_key_capabilities=False
        )  # test may have mangled authorization
        global_b2_tool.cleanup_buckets(buckets_to_clean)


@pytest.fixture(autouse=True, scope='session')
def sample_filepath():
    """Copy the README.md file to /tmp so that docker tests can access it"""
    tmp_readme = pathlib.Path(TEMPDIR) / 'README.md'
    if not tmp_readme.exists():
        tmp_readme.write_text((ROOT_PATH / 'README.md').read_text())
    return tmp_readme


@pytest.fixture(autouse=True, scope='session')
def sample_file(sample_filepath):
    return str(sample_filepath)


@pytest.fixture(scope='session')
def is_running_on_docker(pytestconfig):
    return pytestconfig.getoption('--sut').startswith('docker')


SECRET_FIXTURES = {'application_key', 'application_key_id'}


@pytest.fixture(scope="session")
def homedir(tmp_path_factory):
    yield tmp_path_factory.mktemp("test_homedir")


@pytest.fixture(scope="session")
def b2_in_path(tmp_path_factory):
    """
    Create a dummy b2 executable in a temporary directory and add it to PATH.

    This allows us to test the b2 command from shell level even if tested `b2` package was not installed.
    """

    tempdir = tmp_path_factory.mktemp("temp_bin")
    temp_executable = tempdir / "b2"
    with open(temp_executable, "w") as f:
        f.write(
            f"#!{sys.executable}\n"
            "import sys\n"
            f"sys.path.insert(0, {os.getcwd()!r})\n"  # ensure relative imports work even if command is run in different directory
            "from b2.console_tool import main\n"
            "main()\n"
        )

    temp_executable.chmod(0o700)

    original_path = os.environ["PATH"]
    new_path = f"{tempdir}:{original_path}"
    yield new_path


@pytest.fixture(scope="module")
def env(b2_in_path, homedir, monkeysession, is_running_on_docker):
    """Get ENV for running b2 command from shell level."""
    if not is_running_on_docker:
        monkeysession.setenv('PATH', b2_in_path)
    monkeysession.setenv('HOME', str(homedir))
    monkeysession.setenv('SHELL', "/bin/bash")  # fix for running under github actions
    yield os.environ


@pytest.fixture
def bash_runner(env):
    """Run command in bash shell."""

    def run_command(command: str):
        try:
            return subprocess.run(
                ["/bin/bash", "-c", command],
                capture_output=True,
                check=True,
                env=env,
                text=True,
            )
        except subprocess.CalledProcessError as e:
            print(f"Command {command!r} failed with exit code {e.returncode}")
            print(e.stdout)
            print(e.stderr, file=sys.stderr)
            raise

    return run_command


def pytest_collection_modifyitems(items):
    """
    Add 'require_secrets' marker to all tests that use secrets.
    """
    for item in items:
        if SECRET_FIXTURES & set(getattr(item, 'fixturenames', ())):
            item.add_marker('require_secrets')


@pytest.fixture(scope='module')
def b2_uri_args(apiver_int):
    if apiver_int >= 4:
        return b2_uri_args_v4
    else:
        return b2_uri_args_v3



######################################################################
#
# File: b2/_internal/_cli/arg_parser_types.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import argparse
import functools
import re

import arrow
from b2sdk.v2 import RetentionPeriod

_arrow_version = tuple(int(p) for p in arrow.__version__.split("."))


def parse_comma_separated_list(s):
    """
    Parse comma-separated list.
    """
    return [word.strip() for word in s.split(",")]


def parse_millis_from_float_timestamp(s):
    """
    Parse timestamp, e.g. 1367900664 or 1367900664.152
    """
    parsed = arrow.get(float(s))
    if _arrow_version < (1, 0, 0):
        return int(parsed.format("XSSS"))
    else:
        return int(parsed.format("x")[:13])


def parse_range(s):
    """
    Parse optional integer range
    """
    bytes_range = None
    if s is not None:
        bytes_range = s.split(',')
        if len(bytes_range) != 2:
            raise argparse.ArgumentTypeError('the range must have 2 values: start,end')
        bytes_range = (
            int(bytes_range[0]),
            int(bytes_range[1]),
        )

    return bytes_range


def parse_default_retention_period(s):
    unit_part = '(' + ')|('.join(RetentionPeriod.KNOWN_UNITS) + ')'
    m = re.match(r'^(?P<duration>\d+) (?P<unit>%s)$' % (unit_part), s)
    if not m:
        raise argparse.ArgumentTypeError(
            'default retention period must be in the form of "X days|years "'
        )
    return RetentionPeriod(**{m.group('unit'): int(m.group('duration'))})


def wrap_with_argument_type_error(func, translator=str, exc_type=ValueError):
    """
    Wrap function that may raise an exception into a function that raises ArgumentTypeError error.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exc_type as e:
            raise argparse.ArgumentTypeError(translator(e))

    return wrapper



######################################################################
#
# File: b2/_internal/_utils/uri.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import dataclasses
import pathlib
import urllib.parse
from pathlib import Path
from typing import Sequence

from b2sdk.v2 import (
    B2Api,
    DownloadVersion,
    FileVersion,
    Filter,
)
from b2sdk.v2.exception import B2Error

from b2._internal._utils.python_compat import removeprefix, singledispatchmethod


class B2URIBase:
    pass


@dataclasses.dataclass(frozen=True)
class B2URI(B2URIBase):
    """
    B2 URI designating a particular object by name & bucket or "subdirectory" in a bucket.

    Please note, both files and directories are symbolical concept, not a real one in B2, i.e.
    there is no such thing as "directory" in B2, but it is possible to mimic it by using object names with non-trailing
    slashes.
    To make it possible, it is highly discouraged to use trailing slashes in object names.

    Please note `path` attribute should exclude prefixing slash, i.e. `path` should be empty string for the root of the bucket.
    """

    bucket_name: str
    path: str = ""

    def __str__(self) -> str:
        return f"b2://{self.bucket_name}/{self.path}"

    def is_dir(self) -> bool | None:
        """
        Return if the path is a directory.

        Please note this is symbolical.
        It is possible for file to have a trailing slash, but it is HIGHLY discouraged, and not supported by B2 CLI.
        At the same time it is possible for a directory to not have a trailing slash,
        which is discouraged, but allowed by B2 CLI.
        This is done to mimic unix-like Path.

        In practice, this means that `.is_dir() == True` will always be interpreted as "this is a directory",
        but reverse is not necessary true, and `not uri.is_dir()` should be merely interpreted as
        "this is a directory or a file".

        :return: True if the path is a directory, None if it is unknown
        """
        return not self.path or self.path.endswith("/") or None


@dataclasses.dataclass(frozen=True)
class B2FileIdURI(B2URIBase):
    """
    B2 URI designating a particular file by its id.
    """

    file_id: str

    def __str__(self) -> str:
        return f"b2id://{self.file_id}"


def parse_uri(uri: str, *, allow_all_buckets: bool = False) -> Path | B2URI | B2FileIdURI:
    """
    Parse URI.

    :param uri: string to parse
    :param allow_all_buckets: if True, allow `b2://` without a bucket name to refer to all buckets
    :return: B2 URI or Path
    :raises ValueError: if the URI is invalid
    """
    if not uri:
        raise ValueError("URI cannot be empty")
    parsed = urllib.parse.urlsplit(uri)
    if parsed.scheme == "":
        return pathlib.Path(uri)
    return _parse_b2_uri(uri, parsed, allow_all_buckets=allow_all_buckets)


def parse_b2_uri(
    uri: str, *, allow_all_buckets: bool = False, allow_b2id: bool = True
) -> B2URI | B2FileIdURI:
    """
    Parse B2 URI.

    :param uri: string to parse
    :param allow_all_buckets: if True, allow `b2://` without a bucket name to refer to all buckets
    :param allow_b2id: if True, allow `b2id://` to refer to a file by its id
    :return: B2 URI
    :raises ValueError: if the URI is invalid
    """
    parsed = urllib.parse.urlsplit(uri)
    return _parse_b2_uri(uri, parsed, allow_all_buckets=allow_all_buckets, allow_b2id=allow_b2id)


def _parse_b2_uri(
    uri,
    parsed: urllib.parse.SplitResult,
    *,
    allow_all_buckets: bool = False,
    allow_b2id: bool = True
) -> B2URI | B2FileIdURI:
    if parsed.scheme in ("b2", "b2id"):
        path = urllib.parse.urlunsplit(parsed._replace(scheme="", netloc=""))
        if not parsed.netloc:
            if allow_all_buckets:
                if path:
                    raise ValueError(
                        f"Invalid B2 URI: all buckets URI doesn't allow non-empty path, but {path!r} was provided"
                    )
                return B2URI(bucket_name="")
            raise ValueError(f"Invalid B2 URI: {uri!r}")
        elif parsed.password or parsed.username:
            raise ValueError(
                "Invalid B2 URI: credentials passed using `user@password:` syntax is not supported in URI"
            )

        if parsed.scheme == "b2":
            return B2URI(bucket_name=parsed.netloc, path=removeprefix(path, "/"))
        elif parsed.scheme == "b2id" and allow_b2id:
            return B2FileIdURI(file_id=parsed.netloc)
    else:
        raise ValueError(f"Unsupported URI scheme: {parsed.scheme!r}")


class B2URIAdapter:
    """
    Adapter for using B2URI with B2Api.

    When this matures enough methods from here should be moved to b2sdk B2Api class.
    """

    def __init__(self, api: B2Api):
        self.api = api

    def __getattr__(self, name):
        return getattr(self.api, name)

    @singledispatchmethod
    def download_file_by_uri(self, uri, *args, **kwargs):
        raise NotImplementedError(f"Unsupported URI type: {type(uri)}")

    @download_file_by_uri.register
    def _(self, uri: B2URI, *args, **kwargs):
        bucket = self.get_bucket_by_name(uri.bucket_name)
        return bucket.download_file_by_name(uri.path, *args, **kwargs)

    @download_file_by_uri.register
    def _(self, uri: B2FileIdURI, *args, **kwargs):
        return self.download_file_by_id(uri.file_id, *args, **kwargs)

    @singledispatchmethod
    def get_file_info_by_uri(self, uri, *args, **kwargs):
        raise NotImplementedError(f"Unsupported URI type: {type(uri)}")

    @get_file_info_by_uri.register
    def _(self, uri: B2URI, *args, **kwargs) -> DownloadVersion:
        return self.get_file_info_by_name(uri.bucket_name, uri.path, *args, **kwargs)

    @get_file_info_by_uri.register
    def _(self, uri: B2FileIdURI, *args, **kwargs) -> FileVersion:
        return self.get_file_info(uri.file_id, *args, **kwargs)

    @singledispatchmethod
    def get_download_url_by_uri(self, uri, *args, **kwargs):
        raise NotImplementedError(f"Unsupported URI type: {type(uri)}")

    @get_download_url_by_uri.register
    def _(self, uri: B2URI, *args, **kwargs) -> str:
        return self.get_download_url_for_file_name(uri.bucket_name, uri.path, *args, **kwargs)

    @get_download_url_by_uri.register
    def _(self, uri: B2FileIdURI, *args, **kwargs) -> str:
        return self.get_download_url_for_fileid(uri.file_id, *args, **kwargs)

    @singledispatchmethod
    def list_file_versions_by_uri(self, uri, *args, **kwargs):
        raise NotImplementedError(f"Unsupported URI type: {type(uri)}")

    @list_file_versions_by_uri.register
    def _(self, uri: B2URI, *args, filters: Sequence[Filter] = (), **kwargs):
        bucket = self.api.get_bucket_by_name(uri.bucket_name)
        try:
            yield from bucket.ls(uri.path, *args, filters=filters, **kwargs)
        except ValueError as error:
            # Wrap these errors into B2Error. At the time of writing there's
            # exactly one – `with_wildcard` being passed without `recursive` option.
            raise B2Error(error.args[0])

    @list_file_versions_by_uri.register
    def _(self, uri: B2FileIdURI, *args, **kwargs):
        yield self.get_file_info_by_uri(uri), None



######################################################################
#
# File: test/unit/console_tool/test_authorize_account.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from unittest import mock

import pytest
from b2sdk.v2 import ALL_CAPABILITIES

from b2._internal._cli.const import (
    B2_APPLICATION_KEY_ENV_VAR,
    B2_APPLICATION_KEY_ID_ENV_VAR,
    B2_ENVIRONMENT_ENV_VAR,
)


@pytest.fixture
def b2_cli_is_authorized_afterwards(b2_cli):
    assert b2_cli.account_info.get_account_auth_token() is None
    yield b2_cli
    assert b2_cli.account_info.get_account_auth_token() is not None


def test_authorize_with_bad_key(b2_cli):
    expected_stdout = ""
    expected_stderr = """
    ERROR: unable to authorize account: Invalid authorization token. Server said: secret key is wrong (unauthorized)
    """

    b2_cli._run_command(
        ["authorize-account", b2_cli.account_id, "bad-app-key"],
        expected_stdout,
        expected_stderr,
        1,
    )
    assert b2_cli.account_info.get_account_auth_token() is None


@pytest.mark.parametrize(
    "command",
    [
        "authorize-account",
        "authorize_account",
    ],
)
def test_authorize_with_good_key(b2_cli, b2_cli_is_authorized_afterwards, command):
    assert b2_cli.account_info.get_account_auth_token() is None

    expected_stderr = """
    """

    b2_cli._run_command([command, b2_cli.account_id, b2_cli.master_key], None, expected_stderr, 0)

    assert b2_cli.account_info.get_account_auth_token() is not None


def test_authorize_using_env_variables(b2_cli):
    assert b2_cli.account_info.get_account_auth_token() is None

    expected_stderr = """
    """

    with mock.patch.dict(
        "os.environ",
        {
            B2_APPLICATION_KEY_ID_ENV_VAR: b2_cli.account_id,
            B2_APPLICATION_KEY_ENV_VAR: b2_cli.master_key,
        },
    ):
        b2_cli._run_command(["authorize-account"], None, expected_stderr, 0)

    assert b2_cli.account_info.get_account_auth_token() is not None


@pytest.mark.parametrize(
    "flags,realm_url",
    [
        ([], "http://production.example.com"),
        (["--debug-logs"], "http://production.example.com"),
        (["--environment", "http://custom.example.com"], "http://custom.example.com"),
        (["--environment", "production"], "http://production.example.com"),
        (["--dev"], "http://api.backblazeb2.xyz:8180"),
        (["--staging"], "https://api.backblaze.net"),
    ],
)
def test_authorize_towards_realm(
    b2_cli, b2_cli_is_authorized_afterwards, flags, realm_url, cwd_path, b2_cli_log_fix
):
    expected_stderr = f"Using {realm_url}\n" if any(f != "--debug-logs" for f in flags) else ""

    b2_cli._run_command(
        ["authorize-account", *flags, b2_cli.account_id, b2_cli.master_key],
        None,
        expected_stderr,
        0,
    )
    log_path = cwd_path / "b2_cli.log"
    if "--debug-logs" in flags:
        assert f"Using {realm_url}\n" in log_path.read_text()
    else:
        assert not log_path.exists()


def test_authorize_towards_custom_realm_using_env(b2_cli, b2_cli_is_authorized_afterwards):
    expected_stderr = """
    Using http://custom2.example.com
    """

    with mock.patch.dict(
        "os.environ",
        {
            B2_ENVIRONMENT_ENV_VAR: "http://custom2.example.com",
        },
    ):
        b2_cli._run_command(
            ["authorize-account", b2_cli.account_id, b2_cli.master_key],
            None,
            expected_stderr,
            0,
        )


def test_authorize_account_prints_account_info(b2_cli):
    expected_json = {
        'accountAuthToken': 'auth_token_0',
        'accountFilePath': None,
        'accountId': 'account-0',
        'allowed':
            {
                'bucketId': None,
                'bucketName': None,
                'capabilities': sorted(ALL_CAPABILITIES),
                'namePrefix': None,
            },
        'apiUrl': 'http://api.example.com',
        'applicationKey': 'masterKey-0',
        'applicationKeyId': 'account-0',
        'downloadUrl': 'http://download.example.com',
        'isMasterKey': True,
        's3endpoint': 'http://s3.api.example.com'
    }

    b2_cli._run_command(
        ['authorize-account', b2_cli.account_id, b2_cli.master_key],
        expected_stderr='',
        expected_status=0,
        expected_json_in_stdout=expected_json,
    )



######################################################################
#
# File: test/unit/_cli/test_autocomplete_cache.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# Most of the tests in this module are running in a forked process
# because argcomplete and autocomplete_cache mess with global state,
# making the argument parser unusable for other tests.

from __future__ import annotations

import contextlib
import importlib
import io
import os
import pickle
import sys
from typing import Any

import argcomplete
import pytest

import b2._internal._cli.argcompleters
import b2._internal.arg_parser
import b2._internal.console_tool
from b2._internal._cli import autocomplete_cache

# We can't use pytest.mark.skipif to skip forked tests because with pytest-forked,
# there is an attempt to fork even if the test is marked as skipped.
# See https://github.com/pytest-dev/pytest-forked/issues/44
if sys.platform == "win32":
    forked = pytest.mark.skip(reason="Tests can't be run forked on windows")
else:
    forked = pytest.mark.forked


class Exit:
    """A mocked exit method callable. Instead of actually exiting,
    it just stores the exit code and returns."""

    code: int | None

    @property
    def success(self):
        return self.code == 0

    @property
    def empty(self):
        return self.code is None

    def __init__(self):
        self.code = None

    def __call__(self, n: int):
        self.code = n


@pytest.fixture
def autocomplete_runner(monkeypatch, b2_cli):
    def fdopen(fd, *args, **kwargs):
        # argcomplete package tries to open fd 9 for debugging which causes
        # pytest to later raise errors about bad file descriptors.
        if fd == 9:
            return sys.stderr
        return os.fdopen(fd, *args, **kwargs)

    @contextlib.contextmanager
    def runner(command: str):
        with monkeypatch.context() as m:
            m.setenv('COMP_LINE', command)
            m.setenv('COMP_POINT', str(len(command)))
            m.setenv('_ARGCOMPLETE_IFS', ' ')
            m.setenv('_ARGCOMPLETE', '1')
            m.setattr('os.fdopen', fdopen)

            def _get_b2api_for_profile(profile: str):
                return b2_cli.b2_api

            m.setattr('b2._internal._cli.b2api._get_b2api_for_profile', _get_b2api_for_profile)
            yield

    return runner


def argcomplete_result():
    parser = b2._internal.console_tool.B2.create_parser()
    exit, output = Exit(), io.StringIO()
    argcomplete.autocomplete(parser, exit_method=exit, output_stream=output)
    return exit.code, output.getvalue()


def cached_complete_result(cache: autocomplete_cache.AutocompleteCache, raise_exc: bool = True):
    exit, output = Exit(), io.StringIO()
    cache.autocomplete_from_cache(
        uncached_args={
            'exit_method': exit,
            'output_stream': output
        }, raise_exc=raise_exc
    )
    return exit.code, output.getvalue()


def uncached_complete_result(cache: autocomplete_cache.AutocompleteCache):
    exit, output = Exit(), io.StringIO()
    parser = b2._internal.console_tool.B2.create_parser()
    cache.cache_and_autocomplete(
        parser, uncached_args={
            'exit_method': exit,
            'output_stream': output
        }
    )
    return exit.code, output.getvalue()


@forked
def test_complete_main_command(autocomplete_runner, tmp_path):
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
    )
    with autocomplete_runner('b2 '):
        exit, argcomplete_output = argcomplete_result()
        assert exit == 0
        assert 'get-bucket' in argcomplete_output

    with autocomplete_runner('b2 '):
        exit, output = cached_complete_result(cache)
        # Nothing has been cached yet, we expect simple return, not an exit
        assert exit is None
        assert not output

    with autocomplete_runner('b2 '):
        exit, output = uncached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output

    with autocomplete_runner('b2 '):
        exit, output = cached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output


@forked
def test_complete_with_bucket_suggestions(autocomplete_runner, tmp_path, bucket, authorized_b2_cli):
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
    )
    with autocomplete_runner('b2 get-bucket '):
        exit, argcomplete_output = argcomplete_result()
        assert exit == 0
        assert bucket in argcomplete_output

        exit, output = uncached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output

        exit, output = cached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output


@forked
def test_complete_with_escaped_control_characters(
    autocomplete_runner, tmp_path, bucket, uploaded_file_with_control_chars, authorized_b2_cli
):
    cc_file_name = uploaded_file_with_control_chars['fileName']
    escaped_cc_file_name = uploaded_file_with_control_chars['escapedFileName']
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
    )

    with autocomplete_runner(f'b2 hide-file {bucket} '):
        exit, argcomplete_output = argcomplete_result()
        assert exit == 0
        assert escaped_cc_file_name in argcomplete_output
        assert cc_file_name not in argcomplete_output

        exit, output = uncached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output

        exit, output = cached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output


@forked
def test_complete_with_file_suggestions(
    autocomplete_runner, tmp_path, bucket, uploaded_file, authorized_b2_cli
):
    file_name = uploaded_file['fileName']
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
    )
    with autocomplete_runner(f'b2 hide-file {bucket} '):
        exit, argcomplete_output = argcomplete_result()
        assert exit == 0
        assert file_name in argcomplete_output

        exit, output = cached_complete_result(cache)
        assert exit is None
        assert output == ''

        exit, output = uncached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output

        exit, output = cached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output


@forked
def test_complete_with_file_uri_suggestions(
    autocomplete_runner, tmp_path, bucket, uploaded_file, authorized_b2_cli
):
    file_name = uploaded_file['fileName']
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
    )
    with autocomplete_runner(f'b2 download-file b2://{bucket}/'):
        exit, argcomplete_output = argcomplete_result()
        assert exit == 0
        assert file_name in argcomplete_output

        exit, output = uncached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output

        exit, output = cached_complete_result(cache)
        assert exit == 0
        assert output == argcomplete_output


def test_pickle_store(tmp_path):
    dir = tmp_path
    store = autocomplete_cache.HomeCachePickleStore(dir)

    store.set_pickle('test_1', b'test_data_1')
    assert store.get_pickle('test_1') == b'test_data_1'
    assert store.get_pickle('test_2') is None
    assert len(list(dir.glob('**'))) == 1

    store.set_pickle('test_2', b'test_data_2')
    assert store.get_pickle('test_2') == b'test_data_2'
    assert store.get_pickle('test_1') is None
    assert len(list(dir.glob('**'))) == 1


class Unpickler(pickle.Unpickler):
    """This Unpickler will raise an exception if loading the pickled object
    imports any b2sdk module."""

    _modules_to_load: set[str]

    def load(self):
        self._modules_to_load = set()

        b2_modules = [module for module in sys.modules if 'b2sdk' in module]
        for key in b2_modules:
            del sys.modules[key]

        result = super().load()

        for module in self._modules_to_load:
            importlib.import_module(module)
            importlib.reload(sys.modules[module])

        if any('b2sdk' in module for module in sys.modules):
            raise RuntimeError("Loading the pickled object imported b2sdk module")
        return result

    def find_class(self, module: str, name: str) -> Any:
        self._modules_to_load.add(module)
        return super().find_class(module, name)


def unpickle(data: bytes) -> Any:
    """Unpickling function that raises RuntimeError if unpickled
    object depends on b2sdk."""
    return Unpickler(io.BytesIO(data)).load()


def test_unpickle():
    """This tests ensures that Unpickler works as expected:
    prevents successful unpickling of objects that depend on loading
    modules from b2sdk."""
    from .fixtures.module_loading_b2sdk import function
    pickled = pickle.dumps(function)
    with pytest.raises(RuntimeError):
        unpickle(pickled)


@forked
def test_that_autocomplete_cache_loading_does_not_load_b2sdk(autocomplete_runner, tmp_path):
    cache = autocomplete_cache.AutocompleteCache(
        tracker=autocomplete_cache.VersionTracker(),
        store=autocomplete_cache.HomeCachePickleStore(tmp_path),
        unpickle=unpickle,  # using our unpickling function that fails if b2sdk is loaded
    )
    with autocomplete_runner('b2 '):
        exit, uncached_output = uncached_complete_result(cache)
        assert exit == 0
        assert 'get-bucket' in uncached_output

        exit, output = cached_complete_result(cache, raise_exc=True)
        assert (exit, output) == (0, uncached_output)



######################################################################
#
# File: b2/_internal/_cli/obj_loads.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import argparse
import io
import json
from typing import TypeVar

from b2sdk.v2 import get_b2sdk_doc_urls

try:
    import pydantic
    from pydantic import TypeAdapter, ValidationError
except ImportError:
    pydantic = None


def convert_error_to_human_readable(validation_exc: ValidationError) -> str:
    buf = io.StringIO()
    for error in validation_exc.errors():
        loc = '.'.join(str(loc) for loc in error['loc'])
        buf.write(f'  In field {loc!r} input was `{error["input"]!r}`, error: {error["msg"]}\n')
    return buf.getvalue()


def describe_type(type_) -> str:
    urls = get_b2sdk_doc_urls(type_)
    if urls:
        url_links = ', '.join(f'{name} <{url}>' for name, url in urls.items())
        return f'{type_.__name__} ({url_links})'
    return type_.__name__


T = TypeVar('T')


def validated_loads(data: str, expected_type: type[T] | None = None) -> T:
    if expected_type is not None and pydantic is not None:
        ta = TypeAdapter(expected_type)
        try:
            val = ta.validate_json(data)
        except ValidationError as e:
            errors = convert_error_to_human_readable(e)
            raise argparse.ArgumentTypeError(
                f'Invalid value inputted, expected {describe_type(expected_type)}, got {data!r}, more detail below:\n{errors}'
            ) from e
    else:
        try:
            val = json.loads(data)
        except json.JSONDecodeError as e:
            raise argparse.ArgumentTypeError(f'{data!r} is not a valid JSON value') from e
    return val



######################################################################
#
# File: b2/_internal/b2v3/registry.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# ruff: noqa: F405
from b2._internal._b2v4.registry import *  # noqa
from b2._internal._cli.b2api import _get_b2api_for_profile
from b2._internal.arg_parser import enable_camel_case_arguments
from .rm import Rm

enable_camel_case_arguments()


class ConsoleTool(ConsoleTool):
    # same as original console tool, but does not use InMemoryAccountInfo and InMemoryCache
    # when auth env vars are used

    @classmethod
    def _initialize_b2_api(cls, args: argparse.Namespace, kwargs: dict) -> B2Api:
        return _get_b2api_for_profile(profile=args.profile, **kwargs)


def main() -> None:
    # this is a copy of v4 `main()` but with custom console tool class

    ct = ConsoleTool(stdout=sys.stdout, stderr=sys.stderr)
    exit_status = ct.run_command(sys.argv)
    logger.info('\\\\ %s %s %s //', SEPARATOR, ('exit=%s' % exit_status).center(8), SEPARATOR)

    # I haven't tracked down the root cause yet, but in Python 2.7, the futures
    # packages is hanging on exit sometimes, waiting for a thread to finish.
    # This happens when using sync to upload files.
    sys.stdout.flush()
    sys.stderr.flush()

    logging.shutdown()

    os._exit(exit_status)


class Ls(B2URIBucketNFolderNameArgMixin, BaseLs):
    """
    {BaseLs}

    Examples

    .. note::

        Note the use of quotes, to ensure that special
        characters are not expanded by the shell.


    List csv and tsv files (in any directory, in the whole bucket):

    .. code-block::

        {NAME} ls --recursive --withWildcard bucketName "*.[ct]sv"


    List all info.txt files from directories `b?`, where `?` is any character:

    .. code-block::

        {NAME} ls --recursive --withWildcard bucketName "b?/info.txt"


    List all pdf files from directories b0 to b9 (including sub-directories):

    .. code-block::

        {NAME} ls --recursive --withWildcard bucketName "b[0-9]/*.pdf"


    List all buckets:

    .. code-block::

        {NAME} ls

    Requires capability:

    - **listFiles**
    - **listBuckets** (if bucket name is not provided)
    """
    ALLOW_ALL_BUCKETS = True


B2.register_subcommand(AuthorizeAccount)
B2.register_subcommand(CancelAllUnfinishedLargeFiles)
B2.register_subcommand(CancelLargeFile)
B2.register_subcommand(ClearAccount)
B2.register_subcommand(CopyFileById)
B2.register_subcommand(CreateBucket)
B2.register_subcommand(CreateKey)
B2.register_subcommand(DeleteBucket)
B2.register_subcommand(DeleteFileVersion)
B2.register_subcommand(DeleteKey)
B2.register_subcommand(DownloadFile)
B2.register_subcommand(DownloadFileById)
B2.register_subcommand(DownloadFileByName)
B2.register_subcommand(Cat)
B2.register_subcommand(GetAccountInfo)
B2.register_subcommand(GetBucket)
B2.register_subcommand(FileInfo)
B2.register_subcommand(GetFileInfo)
B2.register_subcommand(GetDownloadAuth)
B2.register_subcommand(GetDownloadUrlWithAuth)
B2.register_subcommand(HideFile)
B2.register_subcommand(ListBuckets)
B2.register_subcommand(ListKeys)
B2.register_subcommand(ListParts)
B2.register_subcommand(ListUnfinishedLargeFiles)
B2.register_subcommand(Ls)
B2.register_subcommand(Rm)
B2.register_subcommand(GetUrl)
B2.register_subcommand(MakeUrl)
B2.register_subcommand(MakeFriendlyUrl)
B2.register_subcommand(Sync)
B2.register_subcommand(UpdateBucket)
B2.register_subcommand(UploadFile)
B2.register_subcommand(UploadUnboundStream)
B2.register_subcommand(UpdateFileLegalHold)
B2.register_subcommand(UpdateFileRetention)
B2.register_subcommand(ReplicationSetup)
B2.register_subcommand(ReplicationDelete)
B2.register_subcommand(ReplicationPause)
B2.register_subcommand(ReplicationUnpause)
B2.register_subcommand(ReplicationStatus)
B2.register_subcommand(Version)
B2.register_subcommand(License)
B2.register_subcommand(InstallAutocomplete)
B2.register_subcommand(NotificationRules)



######################################################################
#
# File: test/unit/console_tool/test_get_url.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest


@pytest.fixture
def uploaded_file_url(bucket_info, uploaded_file):
    return f"http://download.example.com/file/{bucket_info['bucketName']}/{uploaded_file['fileName']}"


@pytest.fixture
def uploaded_file_url_by_id(uploaded_file):
    return f"http://download.example.com/b2api/v2/b2_download_file_by_id?fileId={uploaded_file['fileId']}"


def test_make_url(b2_cli, uploaded_file, uploaded_file_url_by_id):
    b2_cli.run(
        ["make-url", uploaded_file["fileId"]],
        expected_stdout=f"{uploaded_file_url_by_id}\n",
        expected_stderr='WARNING: make-url command is deprecated. Use get-url instead.\n',
    )


def test_make_friendly_url(b2_cli, bucket, uploaded_file, uploaded_file_url):
    b2_cli.run(
        ["make-friendly-url", bucket, uploaded_file["fileName"]],
        expected_stdout=f"{uploaded_file_url}\n",
        expected_stderr='WARNING: make-friendly-url command is deprecated. Use get-url instead.\n',
    )


def test_get_url__b2_uri(b2_cli, bucket, uploaded_file, uploaded_file_url):
    b2_cli.run(
        [
            "get-url",
            f'b2://{bucket}/{uploaded_file["fileName"]}',
        ],
        expected_stdout=f"{uploaded_file_url}\n",
    )


def test_get_url__b2id_uri(b2_cli, uploaded_file, uploaded_file_url_by_id):
    b2_cli.run(
        ["get-url", f'b2id://{uploaded_file["fileId"]}'],
        expected_stdout=f"{uploaded_file_url_by_id}\n",
    )



######################################################################
#
# File: test/unit/console_tool/test_file_info.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest


@pytest.fixture
def uploaded_download_version(b2_cli, bucket_info, uploaded_file):
    return {
        "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
        "contentType": "b2/x-auto",
        "fileId": uploaded_file["fileId"],
        "fileInfo": {
            "src_last_modified_millis": "1500111222000"
        },
        "fileName": "file1.txt",
        "serverSideEncryption": {
            "mode": "none"
        },
        "size": 11,
        "uploadTimestamp": 5000,
    }


@pytest.fixture
def uploaded_file_version(b2_cli, bucket_info, uploaded_file, uploaded_download_version):
    return {
        **uploaded_download_version,
        "accountId": b2_cli.account_id,
        "action": "upload",
        "bucketId": uploaded_file["bucketId"],
    }


def test_get_file_info(b2_cli, uploaded_file_version):
    b2_cli.run(
        ["get-file-info", uploaded_file_version["fileId"]],
        expected_json_in_stdout=uploaded_file_version,
        expected_stderr='WARNING: get-file-info command is deprecated. Use file-info instead.\n',
    )


def test_file_info__b2_uri(b2_cli, bucket, uploaded_download_version):
    b2_cli.run(
        [
            "file-info",
            f'b2://{bucket}/{uploaded_download_version["fileName"]}',
        ],
        expected_json_in_stdout=uploaded_download_version,
    )


def test_file_info__b2id_uri(b2_cli, uploaded_file_version):
    b2_cli.run(
        ["file-info", f'b2id://{uploaded_file_version["fileId"]}'],
        expected_json_in_stdout=uploaded_file_version,
    )



######################################################################
#
# File: test/integration/test_help.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import platform
import re
import subprocess


def test_help(cli_version):
    p = subprocess.run(
        [cli_version, "--help"],
        check=True,
        capture_output=True,
        text=True,
    )

    # verify help contains apiver binary name
    expected_name = cli_version
    if platform.system() == 'Windows':
        expected_name += '.exe'
    assert re.match(r"^_?b2(v\d+)?(\.exe)?$", expected_name)  # test sanity check
    assert f" {expected_name} <command> --help" in p.stdout



######################################################################
#
# File: b2/_internal/_cli/argcompleters.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import itertools

# We import all the necessary modules lazily in completers in order
# to avoid upfront cost of the imports when argcompleter is used for
# autocompletions.
from itertools import islice


def bucket_name_completer(prefix, parsed_args, **kwargs):
    from b2sdk.v2 import unprintable_to_hex

    from b2._internal._cli.b2api import _get_b2api_for_profile
    api = _get_b2api_for_profile(getattr(parsed_args, 'profile', None))
    res = [
        unprintable_to_hex(bucket_name_alias)
        for bucket_name_alias in itertools.chain.from_iterable(
            (bucket.name, f"b2://{bucket.name}") for bucket in api.list_buckets(use_cache=True)
        )
    ]
    return res


def file_name_completer(prefix, parsed_args, **kwargs):
    """
    Completes file names in a bucket.

    To limit delay & cost only lists files returned from by single call to b2_list_file_names
    """
    from b2sdk.v2 import LIST_FILE_NAMES_MAX_LIMIT, unprintable_to_hex

    from b2._internal._cli.b2api import _get_b2api_for_profile

    api = _get_b2api_for_profile(parsed_args.profile)
    bucket = api.get_bucket_by_name(parsed_args.bucketName)
    file_versions = bucket.ls(
        getattr(parsed_args, 'folderName', None) or '',
        latest_only=True,
        recursive=False,
        fetch_count=LIST_FILE_NAMES_MAX_LIMIT,
    )
    return [
        unprintable_to_hex(folder_name or file_version.file_name)
        for file_version, folder_name in islice(file_versions, LIST_FILE_NAMES_MAX_LIMIT)
    ]


def b2uri_file_completer(prefix: str, parsed_args, **kwargs):
    """
    Complete B2 URI pointing to a file-like object in a bucket.
    """
    from b2sdk.v2 import LIST_FILE_NAMES_MAX_LIMIT, unprintable_to_hex

    from b2._internal._cli.b2api import _get_b2api_for_profile
    from b2._internal._utils.python_compat import removeprefix
    from b2._internal._utils.uri import parse_b2_uri

    api = _get_b2api_for_profile(getattr(parsed_args, 'profile', None))
    if prefix.startswith('b2://'):
        prefix_without_scheme = removeprefix(prefix, 'b2://')
        if '/' not in prefix_without_scheme:
            return [
                f"b2://{unprintable_to_hex(bucket.name)}/"
                for bucket in api.list_buckets(use_cache=True)
            ]

        b2_uri = parse_b2_uri(prefix)
        bucket = api.get_bucket_by_name(b2_uri.bucket_name)
        file_versions = bucket.ls(
            f"{b2_uri.path}*",
            latest_only=True,
            recursive=True,
            fetch_count=LIST_FILE_NAMES_MAX_LIMIT,
            with_wildcard=True,
        )
        return [
            unprintable_to_hex(f"b2://{bucket.name}/{file_version.file_name}")
            for file_version, folder_name in islice(file_versions, LIST_FILE_NAMES_MAX_LIMIT)
            if file_version
        ]
    elif prefix.startswith('b2id://'):
        # listing all files from all buckets is unreasonably expensive
        return ["b2id://"]
    else:
        return [
            "b2://",
            "b2id://",
        ]
